@@ -1,7 +1,7 @@
 /* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-
  *
  * nautilus-burn-recorder.c
- * 
+ *
  * Copyright (C) 2002-2004 Bastien Nocera <hadess@hadess.net>
  * Copyright (C) 2005-2006 William Jon McCann <mccann@jhu.edu>
  *
@@ -25,6 +25,7 @@
 
 #include "config.h"
 
+#include <locale.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -86,11 +87,6 @@ enum {
 
 static int nautilus_burn_recorder_table_signals [LAST_SIGNAL] = { 0 };
 
-static GObjectClass *parent_class = NULL;
-
-static void nautilus_burn_recorder_class_init      (NautilusBurnRecorderClass     *class);
-static void nautilus_burn_recorder_init            (NautilusBurnRecorder          *recorder);
-
 static int  nautilus_burn_recorder_write_cdrecord  (NautilusBurnRecorder          *recorder,
 						    NautilusBurnDrive             *drive,
 						    GList                         *tracks,
@@ -150,12 +146,16 @@ nautilus_burn_recorder_track_free (NautilusBurnRecorderTrack *track)
 }
 
 static gboolean
-can_burn_dvds (NautilusBurnDrive *recorder)
+can_burn_dvds (NautilusBurnDrive *drive)
 {
-	if (!(recorder->type & NAUTILUS_BURN_DRIVE_TYPE_DVD_RW_RECORDER) &&
-	    !(recorder->type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_R_RECORDER) &&
-	    !(recorder->type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_R_DL_RECORDER) &&
-	    !(recorder->type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_RW_RECORDER)) {
+	int type;
+
+	type = nautilus_burn_drive_get_drive_type (drive);
+
+	if (!(type & NAUTILUS_BURN_DRIVE_TYPE_DVD_RW_RECORDER) &&
+	    !(type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_R_RECORDER) &&
+	    !(type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_R_DL_RECORDER) &&
+	    !(type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_RW_RECORDER)) {
 		return FALSE;
 	}
 
@@ -200,7 +200,7 @@ cd_write_needs_growisofs (NautilusBurnDrive    *recorder,
 /**
  * nautilus_burn_recorder_cancel:
  * @recorder: #NautilusBurnRecorder
- * @skip_if_dangerous: 
+ * @skip_if_dangerous:
  *
  * Cancel active writing process.
  *
@@ -387,7 +387,7 @@ cdrecord_stdout_line (NautilusBurnProcess   *process,
 		gint64       this_remain;
 		gint64       total;
 		long         secs;
-				
+
 		if (recorder->priv->tracks_total_bytes > 0) {
 			total = recorder->priv->tracks_total_bytes;
 		} else {
@@ -599,17 +599,21 @@ growisofs_stderr_line (NautilusBurnProcess   *process,
 			       NAUTILUS_BURN_RECORDER_MEDIA_DVD);
 	} else if (strstr (line, "unable to open") != NULL || strstr (line, "unable to stat") != NULL) {
 		/* This fits the "open64" and "open"-like messages */
-		process->last_error = g_strdup (_("The recorder could not be accessed"));
-		process->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+		nautilus_burn_process_set_error (process,
+						 _("The recorder could not be accessed."),
+						 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
 	} else if (strstr (line, "not enough space available") != NULL) {
-		process->last_error = g_strdup (_("Not enough space available on the disc"));
-		process->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+		nautilus_burn_process_set_error (process,
+						 _("Not enough space available on the disc."),
+						 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
 	} else if (strstr (line, "end of user area encountered on this track") != NULL) {
-		process->last_error = g_strdup (_("The files selected did not fit on the CD"));
-		process->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+		nautilus_burn_process_set_error (process,
+						 _("The files selected did not fit on the DVD."),
+						 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
 	} else if (strstr (line, "blocks are free") != NULL) {
-		process->last_error = g_strdup (_("The files selected did not fit on the CD"));
-		process->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+		nautilus_burn_process_set_error (process,
+						 _("The files selected did not fit on the DVD."),
+						 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
 	} else if (strstr (line, "flushing cache") != NULL) {
 		g_signal_emit (recorder,
 			       nautilus_burn_recorder_table_signals [ACTION_CHANGED], 0,
@@ -618,9 +622,17 @@ growisofs_stderr_line (NautilusBurnProcess   *process,
 		if (process->result == 0) {
 			process->result = NAUTILUS_BURN_RECORDER_RESULT_FINISHED;
 		}
+	} else if (strstr (line, "unable to unmount") != NULL) {
+		nautilus_burn_process_set_error (process,
+						 _("The target DVD drive is still in use."),
+						 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
+
 	} else if (strstr (line, ":-(") != NULL || strstr (line, "FATAL") != NULL) {
-		process->last_error = g_strdup (_("Unhandled error, aborting"));
-		process->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+		if (! nautilus_burn_process_get_error (process, NULL, NULL)) {
+			nautilus_burn_process_set_error (process,
+							 _("Unhandled error, aborting"),
+							 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
+		}
 	}
 
 	/* for mkisofs exec */
@@ -682,8 +694,9 @@ cdrecord_stderr_line (NautilusBurnProcess   *process,
 
 		insert_cd_retry (process, !res, FALSE, FALSE);
 	} else if (strstr (line, "This means that we are checking recorded media.") != NULL) {
-		process->last_error = g_strdup (_("The CD has already been recorded"));
-		process->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+		nautilus_burn_process_set_error (process,
+						 _("The CD has already been recorded."),
+						 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
 	} else if (strstr (line, "Cannot blank disk, aborting.") != NULL) {
 		gboolean  res;
 
@@ -700,12 +713,13 @@ cdrecord_stderr_line (NautilusBurnProcess   *process,
 
 		insert_cd_retry (process, !res, FALSE, FALSE);
 	} else if (strstr (line, "Data may not fit on current disk") != NULL) {
-		process->last_error = g_strdup (_("The files selected did not fit on the CD"));
-		/* FIXME should we error out in that case?
-		   process->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR; */
+		nautilus_burn_process_set_error (process,
+						 _("The files selected did not fit on the CD."),
+						 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
 	} else if (strstr (line, "Inappropriate audio coding") != NULL) {
-		process->last_error = g_strdup (_("All audio files must be stereo, 16-bit digital audio with 44100Hz samples"));
-		process->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+		nautilus_burn_process_set_error (process,
+						 _("All audio files must be stereo, 16-bit digital audio with 44100Hz samples."),
+						 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
 	} else if (strstr (line, "cannot write medium - incompatible format") != NULL) {
 		gboolean  res;
 
@@ -727,11 +741,14 @@ cdrecord_stderr_line (NautilusBurnProcess   *process,
 		  Sense flags: Blk 65469 (not valid)
 		*/
 		if (strstr (line, "(not valid)") != NULL) {
-			process->last_error = g_strdup (_("Error while writing to disc.  Try a lower speed."));
-			process->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+			nautilus_burn_process_set_error (process,
+							 _("Error while writing to disc.  Try a lower speed."),
+							 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
 		}
 	} else if (strstr (line, "DMA speed too slow") != NULL) {
-		process->last_error = g_strdup (_("The system is too slow to write the CD at this speed. Try a lower speed."));
+		nautilus_burn_process_set_error (process,
+						 _("The system is too slow to write the CD at this speed. Try a lower speed."),
+						 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
 	}
 
 	return TRUE;
@@ -744,7 +761,7 @@ nautilus_burn_recorder_wait_for_insertion (NautilusBurnRecorder *recorder,
 					   gboolean             *blank)
 {
 	NautilusBurnMediaType type;
-	gint64                media_size;
+	gint64                media_capacity;
 	gboolean              reload;
 	gboolean              is_rewritable;
 	gboolean              can_write;
@@ -761,11 +778,11 @@ nautilus_burn_recorder_wait_for_insertion (NautilusBurnRecorder *recorder,
 		gboolean res;
 		res = nautilus_burn_drive_unmount (drive);
 		if (! res) {
-			g_warning ("Couldn't unmount volume in drive: %s", drive->device);
+			g_warning ("Couldn't unmount volume in drive: %s", nautilus_burn_drive_get_device (drive));
 		}
 	}
 
-	media_size = nautilus_burn_drive_get_media_size (drive);
+	media_capacity = nautilus_burn_drive_get_media_capacity (drive);
 	type = nautilus_burn_drive_get_media_type_full (drive, &is_rewritable, &is_blank, &has_data, &has_audio);
 	can_write = nautilus_burn_drive_media_type_is_writable (type, is_blank);
 
@@ -781,7 +798,7 @@ nautilus_burn_recorder_wait_for_insertion (NautilusBurnRecorder *recorder,
 		gboolean is_mounted;
 		int      res;
 
-		reload = (media_size > 0);
+		reload = (media_capacity > 0);
 		if (type == NAUTILUS_BURN_MEDIA_TYPE_ERROR) {
 			reload = FALSE;
 		}
@@ -876,7 +893,7 @@ nautilus_burn_recorder_write_tracks (NautilusBurnRecorder          *recorder,
 		return NAUTILUS_BURN_RECORDER_RESULT_ERROR;
 	}
 
-	is_locked = nautilus_burn_drive_lock (drive, _("Burning CD"), NULL);
+	is_locked = nautilus_burn_drive_lock (drive, _("Burning CD/DVD"), NULL);
 
 	type = nautilus_burn_recorder_wait_for_insertion (recorder,
 							  drive,
@@ -929,9 +946,11 @@ nautilus_burn_recorder_run_process (NautilusBurnRecorder       *recorder,
 	int                  result;
 	GError              *local_error = NULL;
 	NautilusBurnProcess *process;
+	char                *error_message;
+	int                  error_code;
 
  retry:
-	
+
 	process = nautilus_burn_process_new ();
 	process->debug = recorder->priv->debug;
 
@@ -963,11 +982,18 @@ nautilus_burn_recorder_run_process (NautilusBurnRecorder       *recorder,
 		goto retry;
 	}
 
-	if (process->last_error) {
+	if (local_error != NULL) {
 		g_set_error (error,
 			     NAUTILUS_BURN_RECORDER_ERROR,
 			     NAUTILUS_BURN_RECORDER_ERROR_GENERAL,
-			     process->last_error);
+			     local_error->message);
+		g_error_free (local_error);
+	} else if (nautilus_burn_process_get_error (process, &error_message, &error_code)) {
+		g_set_error (error,
+			     NAUTILUS_BURN_RECORDER_ERROR,
+			     NAUTILUS_BURN_RECORDER_ERROR_GENERAL,
+			     error_message);
+		g_free (error_message);
 	}
 
 	result = process->result;
@@ -1091,9 +1117,20 @@ nautilus_burn_recorder_write_growisofs (NautilusBurnRecorder          *recorder,
 
 	g_ptr_array_add (argv, "growisofs");
 	if (speed != 0) {
-		speed = NAUTILUS_BURN_DRIVE_DVD_SPEED (speed);
+		char	*current_locale = NULL;
+		float	 dvd_speed;
+
+		dvd_speed = NAUTILUS_BURN_DRIVE_DVD_SPEED (speed);
+
+		/* We want g_strdup_printf() to use dots as decimal
+		 * point for the speed, otherwise growisofs wouldn't understand. */
+		current_locale = setlocale (LC_NUMERIC, NULL);
+		setlocale (LC_NUMERIC, "C");
+
+		speed_str = g_strdup_printf ("-speed=%.1f", dvd_speed);
+
+		setlocale (LC_NUMERIC, current_locale);
 
-		speed_str = g_strdup_printf ("-speed=%d", speed);
 		g_ptr_array_add (argv, speed_str);
 	}
 
@@ -1118,10 +1155,10 @@ nautilus_burn_recorder_write_growisofs (NautilusBurnRecorder          *recorder,
 
 	dev_str = NULL;
 	if (t->type == NAUTILUS_BURN_RECORDER_TRACK_TYPE_DATA) {
-		dev_str = g_strdup_printf ("%s=%s", drive->device, t->contents.data.filename);
+		dev_str = g_strdup_printf ("%s=%s", nautilus_burn_drive_get_device (drive), t->contents.data.filename);
 		g_ptr_array_add (argv, dev_str);
 	} else {
-		g_ptr_array_add (argv, drive->device);
+		g_ptr_array_add (argv, (char *)nautilus_burn_drive_get_device (drive));
 
 		/* mkisofs options */
 		if (ncb_mkisofs_supports_utf8 ()) {
@@ -1231,6 +1268,12 @@ cdrdao_stderr_line (NautilusBurnProcess   *process,
 
 	}
 
+	if (strstr (line, "Operation not permitted")) {
+		nautilus_burn_process_set_error (process,
+						 line,
+						 NAUTILUS_BURN_RECORDER_RESULT_ERROR);
+	}
+
 	if (strstr (line, "Writing finished successfully")) {
 		process->dangerous = FALSE;
 		process->result = NAUTILUS_BURN_RECORDER_RESULT_FINISHED;
@@ -1270,10 +1313,13 @@ nautilus_burn_recorder_write_cdrecord (NautilusBurnRecorder          *recorder,
 		int                            res;
 		GError                        *err = NULL;
 
-		if (flags & NAUTILUS_BURN_RECORDER_WRITE_DEBUG)
+		if (flags & NAUTILUS_BURN_RECORDER_WRITE_DEBUG) {
 			blank_flags |= NAUTILUS_BURN_RECORDER_BLANK_DEBUG;
-		if (flags & NAUTILUS_BURN_RECORDER_WRITE_DUMMY_WRITE)
+		}
+		if (flags & NAUTILUS_BURN_RECORDER_WRITE_DUMMY_WRITE) {
 			blank_flags |= NAUTILUS_BURN_RECORDER_BLANK_DUMMY_WRITE;
+		}
+
 		/* cdrdao doesn't seem to have on-the-fly blank */
 		/* cdrecord doesn't seem to correctly detect disc size when blanking */
 		res = nautilus_burn_recorder_blank_disc (recorder,
@@ -1281,18 +1327,25 @@ nautilus_burn_recorder_write_cdrecord (NautilusBurnRecorder          *recorder,
 							 NAUTILUS_BURN_RECORDER_BLANK_FAST,
 							 blank_flags,
 							 &err);
-		if (res != NAUTILUS_BURN_RECORDER_RESULT_FINISHED)
+		if (err != NULL) {
+			g_propagate_error (error, err);
+		}
+
+		if (res != NAUTILUS_BURN_RECORDER_RESULT_FINISHED) {
 			return res;
+		}
 	}
 
-	speed = NAUTILUS_BURN_DRIVE_CD_SPEED (speed);
+	/* Turn the kB/s to a CD speed (for example 10 for "10x").
+	 * cdrecord supports only integer numbers for the speed. */
+	 speed = (int) NAUTILUS_BURN_DRIVE_CD_SPEED (speed);
 
 	if (track->type ==  NAUTILUS_BURN_RECORDER_TRACK_TYPE_CUE) {
 		argv = g_ptr_array_new ();
 		g_ptr_array_add (argv, "cdrdao");
 		g_ptr_array_add (argv, "write");
 		g_ptr_array_add (argv, "--device");
-		g_ptr_array_add (argv, drive->cdrecord_id);
+		g_ptr_array_add (argv, (char *)nautilus_burn_drive_get_device (drive));
 		g_ptr_array_add (argv, "--speed");
 		speed_str = g_strdup_printf ("%d", speed);
 		g_ptr_array_add (argv, speed_str);
@@ -1324,7 +1377,7 @@ nautilus_burn_recorder_write_cdrecord (NautilusBurnRecorder          *recorder,
 		if (speed != 0) {
 			g_ptr_array_add (argv, speed_str);
 		}
-		dev_str = g_strdup_printf ("dev=%s", drive->cdrecord_id);
+		dev_str = g_strdup_printf ("dev=%s", nautilus_burn_drive_get_device (drive));
 		g_ptr_array_add (argv, dev_str);
 		if (flags & NAUTILUS_BURN_RECORDER_WRITE_DUMMY_WRITE) {
 			g_ptr_array_add (argv, "-dummy");
@@ -1377,13 +1430,18 @@ nautilus_burn_recorder_write_cdrecord (NautilusBurnRecorder          *recorder,
 		err_line_func = cdrecord_stderr_line;
 	}
 
-	if (argv) {
+	if (argv != NULL) {
+		GError *err = NULL;
+
 		result = nautilus_burn_recorder_run_process (recorder,
 							     NAUTILUS_BURN_RECORDER_MEDIA_CD,
 							     argv,
 							     out_line_func,
 							     err_line_func,
-							     error);
+							     &err);
+		if (err != NULL) {
+			g_propagate_error (error, err);
+		}
 	}
 
 	g_free (cue_str);
@@ -1410,7 +1468,7 @@ nautilus_burn_recorder_blank_disc_cdrecord (NautilusBurnRecorder          *recor
 	gboolean              is_blank;
 	int                   result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
 
-	recorder->priv->can_rewrite = (drive->type & NAUTILUS_BURN_DRIVE_TYPE_CDRW_RECORDER);
+	recorder->priv->can_rewrite = nautilus_burn_drive_can_rewrite (drive);
 
 	if (!recorder->priv->can_rewrite) {
 		return NAUTILUS_BURN_RECORDER_RESULT_CANCEL;
@@ -1430,7 +1488,7 @@ nautilus_burn_recorder_blank_disc_cdrecord (NautilusBurnRecorder          *recor
 	argv = g_ptr_array_new ();
 	g_ptr_array_add (argv, "cdrecord");
 
-	dev_str = g_strdup_printf ("dev=%s", drive->cdrecord_id);
+	dev_str = g_strdup_printf ("dev=%s", nautilus_burn_drive_get_device (drive));
 	g_ptr_array_add (argv, dev_str);
 	g_ptr_array_add (argv, "-v");
 
@@ -1458,7 +1516,7 @@ nautilus_burn_recorder_blank_disc_cdrecord (NautilusBurnRecorder          *recor
 	g_free (dev_str);
 	g_free (blank_str);
 	g_ptr_array_free (argv, TRUE);
-	
+
 	return result;
 }
 
@@ -1480,7 +1538,7 @@ nautilus_burn_recorder_blank_disc_dvdrw_format (NautilusBurnRecorder          *r
 	gboolean                    is_blank;
 	int                         result         = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
 
-	recorder->priv->can_rewrite = (drive->type & NAUTILUS_BURN_DRIVE_TYPE_CDRW_RECORDER);
+	recorder->priv->can_rewrite = nautilus_burn_drive_can_rewrite (drive);
 
 	if (!recorder->priv->can_rewrite) {
 		return NAUTILUS_BURN_RECORDER_RESULT_CANCEL;
@@ -1504,7 +1562,7 @@ nautilus_burn_recorder_blank_disc_dvdrw_format (NautilusBurnRecorder          *r
 		g_ptr_array_add (argv, "growisofs");
 		g_ptr_array_add (argv, "-Z");
 
-		dev_str = g_strdup_printf ("%s=%s", drive->device, "/dev/zero");
+		dev_str = g_strdup_printf ("%s=%s", nautilus_burn_drive_get_device (drive), "/dev/zero");
 		g_ptr_array_add (argv, dev_str);
 		g_ptr_array_add (argv, NULL);
 
@@ -1527,7 +1585,7 @@ nautilus_burn_recorder_blank_disc_dvdrw_format (NautilusBurnRecorder          *r
 			g_ptr_array_add (argv, blank_str);
 		}
 
-		dev_str = g_strdup_printf ("%s", drive->device);
+		dev_str = g_strdup_printf ("%s", nautilus_burn_drive_get_device (drive));
 		g_ptr_array_add (argv, dev_str);
 		g_ptr_array_add (argv, NULL);
 
@@ -1549,11 +1607,10 @@ nautilus_burn_recorder_blank_disc_dvdrw_format (NautilusBurnRecorder          *r
 	if (is_locked) {
 		nautilus_burn_drive_unlock (drive);
 	}
-	
+
 	return result;
 }
 
-
 static gboolean
 nautilus_burn_drive_format_needs_growisofs (NautilusBurnDrive    *drive,
 					    NautilusBurnMediaType type)
@@ -1622,8 +1679,8 @@ nautilus_burn_recorder_finalize (GObject *object)
 		nautilus_burn_process_free (recorder->priv->process);
 	}
 
-	if (G_OBJECT_CLASS (parent_class)->finalize != NULL) {
-		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
+	if (G_OBJECT_CLASS (nautilus_burn_recorder_parent_class)->finalize != NULL) {
+		(* G_OBJECT_CLASS (nautilus_burn_recorder_parent_class)->finalize) (object);
 	}
 }
 
@@ -1653,8 +1710,6 @@ nautilus_burn_recorder_class_init (NautilusBurnRecorderClass *klass)
 
 	object_class = (GObjectClass *) klass;
 
-	parent_class = g_type_class_peek_parent (klass);
-
 	object_class->finalize = nautilus_burn_recorder_finalize;
 
 	g_type_class_add_private (klass, sizeof (NautilusBurnRecorderPrivate));
